/* BMP to ASCII converter, ver 1.1, 1994
   synopsis:
   bmp2asc <bitmap> [<width> [<vertical_shrink> [<outfile>]]]
              <bitmap>  Windows BMP monochrome btimap file name
               <width>  width of output in characters
     <vertical_shrink>  specifies output height in percents of original.
                        default is 50. This is to compensate fact that
                        letters aren't square
             <outfile>  name of target file.

   comments, bug_reports, suggestions, fuckings etc goes to:

                              ex599@cleveland.freenet.edu

                                have fun!
   */

#ifndef __COMPACT__
#error This program is written to be compiled only with compact memory model
#endif

#include <stdio.h>
#include <alloc.h>
#include <string.h>
#include <conio.h>
#include <ctype.h>


// global constants
#define OUTPAL_SIZE       0x40
#define OUTPAL_FACTOR     (0x100 / OUTPAL_SIZE)
#define OUTPAL_UNUSED     '_'
#define DEF_WIDTH         67
#define DEF_VERT_FACTOR   50


// Error Codes - BMP io   (0x01 - 0x0F)
#define ERR_NOTBMP          0x01
#define ERR_NOTWINBMP       0x02
#define ERR_BMPCOMPRESSED   0x03
#define ERR_BMPPLANES       0x04
#define ERR_BMPNOTMONO      0x05
// Error Codes - memory allocation (0x10 - 0x1F)
#define ERR_OUTMEM          0x10
// Error Codes - file io (0x20 - 0x2F)
#define ERR_IOGENERIC       0x20
#define ERR_IOOPEN          0x21
#define ERR_IOREAD          0x22
#define ERR_IOSEEK          0x23
#define ERR_IOWRITE         0x24
// Error Codes - integrity and bugs (0x30 - 0x4F)
#define ERR_HEADER          0x30
#define ERR_COORDINATES     0x31


// Keystrokes
#define CtrlLeft            115
#define CtrlRight           116
#define CtrlUp              141
#define CtrlDown            145
#define Up                  72
#define Down                80
#define Left                75
#define Right               77
#define Enter               '\r'
#define Esc                 27
#define CtrlP               16
#define CtrlE               5
#define MINCOLOR            0x20
#define MAXCOLOR            0x7F


int errornum;  // here are error codes placed by functions that can't
               // return them explicitly
char default_palette[OUTPAL_SIZE + 1] = "\
            ::::::::::::oooooooooooooooommmmmmmmmmmm@@@@@@@@@@@@";
         // default output palette.


/* BMP structure is for keeping BITMAP parameters */
typedef struct tagBMP {
  FILE *bmpfile;                // pointer to stream assigned to the file
  long picturedata;             // offset of picture data in BMP file
  long width;                   // width of picture in pixels
  long height;                  // height of picture in pixels
  long linelen;                 // lenght of picture line in bytes
  int bpp;                      // bits per pixel count
  unsigned char huge *buffer;   // pointer to image data (set by ReadBitmap)
  } BMP;

/* OUTP structure for keeping output parameters */
typedef struct tagOUTP {
  long x, y;                   // sizes of output picture (columns by rows)
  long dx100;                  // sizes of rectangle of input pixels that
  long dy100;                  // form one output pixel (multiplied by 100)
  unsigned char huge *buffer;  // pointer to output image data
  char palette[OUTPAL_SIZE + 1];   // output palette charcters
  } OUTP;



/* ReadBMPHeader()  reads header and fills passed BMP sturcture.
                    Returns:
                    0                   success
                    ERR_NOTBMP          not BMP file
                    ERR_NOTWINBMP       not Windows BMP file
                    ERR_BMPCOMPRESSED   bitmap is compressed
                    ERR_BMPPLANES       bitmap has multiple planes
                    ERR_BMPNOTMONO      not monochrome bitmap
                    ERR_IO...           errors when reading the file
*/
int ReadBMPHeader(BMP *header, char *filename)
{
long test = 0;
if ((header->bmpfile = fopen(filename, "rb")) == NULL) return ERR_IOOPEN;
if (fseek(header->bmpfile, 0, SEEK_SET) != 0) return ERR_IOSEEK;
if (fread(&test, 2, 1, header->bmpfile) < 1) return ERR_IOREAD;
if (test != 'B' + 'M' * 256) return ERR_NOTBMP;
if (fseek(header->bmpfile, 10, SEEK_SET) != 0) return ERR_IOSEEK;
if (fread(&(header->picturedata), 4, 1, header->bmpfile) < 1) return ERR_IOREAD;
if (fread(&test, 4, 1, header->bmpfile) < 1) return ERR_IOREAD;
if (test != 40) return ERR_NOTWINBMP;
if (fread(&(header->width), 4, 1, header->bmpfile) < 1) return ERR_IOREAD;
if (fread(&(header->height), 4, 1, header->bmpfile) < 1) return ERR_IOREAD;
test = 0;
if (fread(&test, 2, 1, header->bmpfile) < 1) return ERR_IOREAD;
if (test != 1) return ERR_BMPPLANES;
if (fread(&(header->bpp), 2, 1, header->bmpfile) < 1) return ERR_IOREAD;
if (header->bpp != 1) return ERR_BMPNOTMONO;
if (fread(&test, 4, 1, header->bmpfile) < 1) return ERR_IOREAD;
if (test != 0) return ERR_BMPCOMPRESSED;
header->linelen = ((header->width / 32) + (header->width % 32 != 0)) * 4;
return 0;
}


/* ReadBitmap  reads picture data into memory, doing appropriate
               conversion: picture is no longer flipped upside
               down, that is the greater y coordinate (vertical), the
               greater the addres in memory
               Returns:
               0              succsess;
               ERR_OUTMEM     failed to allocate memory
               ERR_IO...      errors when reading the file
               ERR_HEADER     invalid header passed
*/
int ReadBitmap(BMP *header)
{
unsigned long imagesize, y;
switch (header->bpp) {
  case 1: imagesize = header->linelen * header->height;
          break;
  default: return ERR_HEADER;
  }
if ((header->buffer = (unsigned char huge *)farmalloc(imagesize)) == NULL)
  return ERR_OUTMEM;
if (fseek(header->bmpfile, header->picturedata, SEEK_SET) != 0)
  return ERR_IOSEEK;
for (y = header->height; y > 0; y--) {
  if (fread(header->buffer + (y - 1) * header->linelen,
            header->linelen, 1, header->bmpfile) < 1)
    return ERR_IOREAD;
  }
return 0;
}


/* ReadInPixel  returns bitmap pixel value, i.e. idex to the
                array of colors. Returns
                -1           error ocured (look up variable errornum for error
                                           code)
                0 or positive the pixel value, success.
                errornum settings:
                ERR_HEADER        wrong header passed
                ERR_COORDINATES   to big coordinate passed
                ERR_MEMCORRUPTED  well...
*/
int ReadInPixel(BMP *header, unsigned long px, unsigned long py)
{
unsigned char pixbyte,pixbit,pix;
switch (header->bpp) {
  case 1:  break;
  default: errornum = ERR_HEADER;
           return -1;
  }
if (px >= header->width || py >= header->height) {
  errornum = ERR_COORDINATES;
  return -1;
  }
switch (header->bpp) {
  case 1: pixbyte = header->buffer[header->linelen * py + (px / 8)];
          pixbit = 7 - (px % 8);
          pix = (pixbyte >> pixbit) & 0x01;
          break;
  default: errornum = ERR_HEADER;  // this is unreachable code, but if I forget to
           return -1;              // modify it in future versions...
  }
return pix;
}


/* CalcOutParams   calculates parameters used to create output image.
                   Returns:
                   0            success
                   ERR_OUTMEM   parameters calculated, but no memory
                                allocated for the output image
                   ERR_HEADER   wrong BMP header passed, no calculations
                                performed
*/
int CalcOutParams(BMP *header, OUTP *param, int x, int vsp)
{
if (header->height == 0 || header->width == 0) return ERR_HEADER;
param->x = x;
param->y = param->x * header->height / header->width * vsp / 100;
param->y += (param->y == 0);
param->dx100 = header->width * 100 / param->x;
param->dy100 = header->height * 100 / param->y;
_fstrcpy(param->palette, default_palette);
if ((param->buffer = (unsigned char huge *)farmalloc(param->x * param->y)) == NULL)
  return ERR_OUTMEM;
return 0;
}


/* FormOutPixel   reads appropraite input pixels and translates them
                  into output pixel. ( 0 .. 255 color value)
                  Returns:
                  0 to 255  success.
                  -1  error occured, look up errornum for error code
*/
int FormOutPixel(BMP *header, OUTP *param, long x, long y)
{
long x1, x2, y1, y2, i, j;
long prepix, pix, npixels;
x1 = x * param->dx100 / 100;            // calculate input rectangle that
x2 = (x + 1) * param->dx100 / 100;      // determines color of output pixel
y1 = y * param->dy100 / 100;
y2 = (y + 1) * param->dy100 / 100;
if (x2 > header->width) x2 = header->width;
if (y2 > header->height) y2 = header->height;

i = x1;                            // conversion loop
npixels = pix = 0;
do {
  j = y1;
  do {
    if ((prepix = ReadInPixel(header, i, j)) == -1)
      return -1;           // pass on error code from ReadInPixel
    pix += 255 * prepix;
    npixels++;
    j++;
    } while (j<y2);
  i++;
  } while (i<x2);
return (pix / npixels);  // return value ranging from 0 to 255
}


/* Convert  converts input picture to intermediate representation
            Returns:
            0  success
            appropriate error codes from called functions if they fail
*/
int Convert(BMP *header, OUTP *param)
{
long x, y;
int pix;
for (y = 0; y < param->y; y++) {
  for (x = 0; x < param->x; x++) {
    if ((pix = FormOutPixel(header, param, x, y)) == -1) return errornum;
    param->buffer[param->x * y + x] = pix / OUTPAL_FACTOR;
    }
  }
return 0;
}



/* CleanPalette       Searches image and builds histogram of used colors
                      then puts OUTPAL_UNUSED characters into palette entries
                      that are unused
                      */
void CleanPalette(OUTP *param)
{
long x,y;
int hist[OUTPAL_SIZE];
for (x = 0; x < OUTPAL_SIZE; x++) hist[x] = 0;
for (y = 0; y < param->y; y++)
  for (x = 0; x < param->x; x++)
    hist[param->buffer[param->x * y + x]]++;
for (x = 0; x < OUTPAL_SIZE; x++) {
  if (hist[x]) continue;
  param->palette[x] = OUTPAL_UNUSED;
  }
}

/* Edit     allaws to make retouch - you can assign to the pixel another
            palette entry
            */
void Edit(OUTP *param, int x, int y)
{
int curx, cury, i, j;
int palx;
int refresh = 1, done = 0;
int c;
unsigned char atr;
struct text_info t;
gettextinfo(&t);
atr = t.attribute;
curx = x;
cury = y;
_wscroll = 0;
do {
  palx = param->buffer[y * param->x + x];
  switch (refresh) {
    case 0: break;
    case 1: _setcursortype(_NOCURSOR);
            gotoxy(1,1);
            for (j = cury; j < cury + 24 && j < param->y; j++) {
              for (i = curx; i < curx + 80 && i < param->x; i++)
                putch(param->palette[param->buffer[param->x * j + i]]);
              cputs("\r\n");
              }
            textattr(((atr&0x88)|((atr>>4)&0x07)|((atr<<4)&0x70))^0x08);
            gotoxy(1,25);
            cputs(param->palette);
            cprintf(" x=%-3d y=%-3d", x, y);
            clreol();
            textattr(atr);
            gotoxy(palx + 1, 25);
            putch(param->palette[palx]);
            gotoxy(x - curx + 1, y - cury + 1);
            _setcursortype(_NORMALCURSOR);
            refresh = 0;
            break;
    case 2: _setcursortype(_NOCURSOR);
            textattr(((atr&0x88)|((atr>>4)&0x07)|((atr<<4)&0x70))^0x08);
            gotoxy(1,25);
            cputs(param->palette);
            cprintf(" x=%-3d y=%-3d", x, y);
            clreol();
            textattr(atr);
            gotoxy(palx + 1, 25);
            putch(param->palette[palx]);
            gotoxy(x - curx + 1, y - cury + 1);
            putch(param->palette[param->buffer[param->x * y + x]]);
            gotoxy(x - curx + 1, y - cury + 1);
            _setcursortype(_NORMALCURSOR);
            break;
    default: break;
    }
  switch (c = getch()) {
    case 0:
      switch (getch()) {
        case CtrlLeft:   if (x > 0)
                           { x--; refresh = 2; }
                         if (curx > x)
                           { curx = x; refresh = 1; }
                         break;
        case CtrlRight:  if (x < param->x - 1)
                           { x++; refresh = 2; }
                         if (curx < x - 79)
                           { curx = x - 79; refresh = 1; }
                         break;
        case Up:         ;
        case CtrlUp:     if (y > 0)
                           { y--; refresh = 2; }
                         if (cury > y)
                           { cury = y; refresh = 1; }
                         break;
        case Down:       ;
        case CtrlDown:   if (y < param->y - 1)
                           { y++; refresh = 2; }
                         if (cury < y - 23)
                           { cury = y - 23; refresh = 1;}
                         break;
        case Left:       palx -= palx > 0;
                         param->buffer[y * param->x + x] = palx;
                         refresh = 2;
                         break;
        case Right:      palx += palx < OUTPAL_SIZE -1;
                         param->buffer[y * param->x + x] = palx;
                         refresh = 2;
                         break;
        default:         break;
        } // end of inner switch
      break; // end of case 0 in outer switch
    case Esc: done = 1;
              break;
    case CtrlP: CleanPalette(param);
                refresh = 1;
                break;
    default:
      if (c < MINCOLOR || c > MAXCOLOR) ;
      else {
        param->palette[palx] = c;
        refresh = 1;
        }
      break;
    } // end of outer switch
  } while (!done);
}



/* Preview    allows you to preview converted image, and adjust
              output palette.
              Returns:
               0: file is to be saved
               1: abort save
*/
int Preview(OUTP *param)
{
int curx = 0, cury = 0, i, j;
int palx = 0;
int refresh = 1, done = 0, save = 1;
int c;
unsigned char atr;
struct text_info t;
clrscr();
gettextinfo(&t);
atr = t.attribute;
_wscroll = 0;
do {
  if (refresh) {
    _setcursortype(_NOCURSOR);
    gotoxy(1,1);
    for (j = cury; j < cury + 24 && j < param->y; j++) {
      for (i = curx; i < curx + 80 && i < param->x; i++)
        putch(param->palette[param->buffer[param->x * j + i]]);
      cputs("\r\n");
      }
    textattr(((atr&0x88)|((atr>>4)&0x07)|((atr<<4)&0x70))^0x08);
    gotoxy(1,25);
    cputs(param->palette);
    cprintf(" x=%-3d y=%-3d", curx, cury);
    clreol();
    textattr(atr);
    gotoxy(palx + 1, 25);
    _setcursortype(_NORMALCURSOR);
    refresh = 0;
    }  // end of if (refersh)
  switch (c = getch()) {
    case 0:
      switch (getch()) {
        case CtrlLeft:   if (curx > 0)
                           { curx--; refresh = 1; }
                         break;
        case CtrlRight:  if (curx < param->x - 80)
                           { curx++; refresh = 1; }
                         break;
        case Up:         ;
        case CtrlUp:     if (cury > 0)
                           { cury--; refresh = 1; }
                         break;
        case Down:       ;
        case CtrlDown:   if (cury < param->y - 24)
                           { cury++; refresh = 1; }
                         break;
        case Left:       palx -= palx > 0;
                         gotoxy(palx + 1, 25);
                         break;
        case Right:      palx += palx < OUTPAL_SIZE -1;
                         gotoxy(palx + 1, 25);
                         break;
        default:         break;
        } // end of inner switch
      break; // end of case 0 in outer switch
    case Enter: done = 1;
                break;
    case Esc: done = 1;
              save = 0;
              break;
    case CtrlP: CleanPalette(param);
                refresh = 1;
                break;
    case CtrlE: Edit(param, curx, cury);
                refresh = 1;
                break;
    default:
      if (c < MINCOLOR || c > MAXCOLOR) ;
      else {
        param->palette[palx] = c;
        palx += palx < OUTPAL_SIZE - 1;
        refresh = 1;
        }
      break;
    } // end of outer switch
  } while (!done);
_wscroll = 1;
clrscr();
return save;
}


/* Save    saves output image to a file, replacing indexes to a palette
           with palette values.
           Returns:
           0             success
           ERR_IOOPEN      error opening file
           ERR_IOWRITE     error writing file
*/
int Save(OUTP *param, char *filename)
{
FILE *outfile;
long x, y;
int c;
if ((outfile = fopen(filename, "wt")) == NULL) return ERR_IOOPEN;
for (y = 0; y < param->y; y++) {
  for (x = 0; x < param->x; x++) {
    c = param->palette[param->buffer[param->x * y + x]];
    if (fputc(c,outfile) == EOF) return ERR_IOWRITE;
    }
  if (fputc('\n',outfile) == EOF) return ERR_IOWRITE;
  }
fclose(outfile);
return 0;
}


/* ================================================================
   ==========================   MAIN   ============================
   ================================================================ */

char *title = "BMP to ASCII converter 1.1 by brojek\n";
char *usage = "Usage:\n\
  bmp2asc <bmpfile> [<width> [<vertical_factor> [<outfile>]]]\n\
\n\
          <bmpfile>  name of BMP file to convert. (extension not necessary)\n\
            <width>  width of output in characters.\n\
  <vertical_factor>  percentage of original height. use it if picture looks\n\
                     \"tall\" or \"short\"\n\
          <outfile>  name of output file.\n\
\n\
Please direct bugs reports, comments or suggestions to\n\
  ex599@cleveland.freenet.edu\n\
Include \"bmp2asc\" in subject of the message.\n";


/* FormatFileName   adds a given extension to the filename if it hasn't
                    an extension yet */
void FormatFileName(char *str, char *ext)
{
int i=0;
while (str[i]!='\0') {if (islower(str[i])) str[i]=_toupper(str[i]); i++;}
if (str[i-1]=='.') { str[i-1]='\0'; return; }
if (str[i-2]=='.' || str[i-3]=='.' || str[i-4]=='.') return;
_fstrcpy(str + i, ext);
}

/* RemoveExtension   removes extension from the filename */
void RemoveExtension(char *str)
{
int i=0, j;
while (str[i]!='\0') {if (islower(str[i])) str[i]=_toupper(str[i]); i++;}
for (j=1; j<=4; j++) {
  if (str[i-j]=='.') { str[i-j]='\0'; return; }
  }
}

/* IoErrorReport   obvious */
void IoErrorReport(int err)
{
fputs("Error ",stderr);
switch (err) {
  case ERR_IOOPEN: fputs("opening", stderr); break;
  case ERR_IOSEEK: fputs("seeking", stderr); break;
  case ERR_IOREAD: fputs("reading", stderr); break;
  case ERR_IOWRITE: fputs("writing", stderr); break;
  default: fputs("<unknown> working with", stderr); break;
  }
fputs(" file.\n", stderr);
}



int main(int argc, char *argv[])
{
BMP header;
OUTP param;
int err;
int width, vsf;
char infile[80], outfile[80];
fputs(title,stderr);
switch (argc) {
  case 5: _fstrcpy(outfile, argv[4]);
  case 4: sscanf(argv[3], "%Fu", &vsf);
  case 3: sscanf(argv[2], "%Fu", &width);
  case 2: _fstrcpy(infile, argv[1]);
          break;
  default: fputs(usage,stderr);
           return 1;
  }
if (argc < 5) {
  strcpy(outfile, infile);
  RemoveExtension(outfile);
  FormatFileName(outfile,".TXT");
  }
if (argc < 4) vsf = DEF_VERT_FACTOR;
if (argc < 3) width = DEF_WIDTH;
FormatFileName(infile,".BMP");
fprintf(stderr,"Reading header of input file: %s\n",infile);
switch (err = ReadBMPHeader(&header, infile)) {
  case 0: break;
  case ERR_NOTBMP: fputs("Not BMP file.\n",stderr);
                   return 2;
  case ERR_NOTWINBMP: fputs("Not Windows BMP file.\n",stderr);
                      return 2;
  case ERR_BMPCOMPRESSED: fputs("This Bitmap is compressed.\n", stderr);
                          return 2;
  case ERR_BMPPLANES: fputs("This bitmap has multiple planes.\n", stderr);
                      return 2;
  case ERR_BMPNOTMONO: fputs("Not monochrome bitmap.\n", stderr);
                       return 2;
  case ERR_IOOPEN: ;
  case ERR_IOSEEK: ;
  case ERR_IOREAD: IoErrorReport(err);
                   return 3;
  default: fputs("Unknown error occured == BUG in program.\n", stderr);
           return 100;
  }
fputs("Reading bitmap.\n", stderr);
switch (err = ReadBitmap(&header)) {
   case 0: break;
   case ERR_OUTMEM: fputs("Failed to allocate memory for the bitmap.\n", stderr);
                    return 4;
   case ERR_HEADER: fputs("Invalid header passed.\n", stderr);
                    return 2;
   case ERR_IOSEEK: ;
   case ERR_IOREAD: IoErrorReport(err);
                    return 3;
   default: fputs("Unknown error occured == BUG in program.\n", stderr);
            return 100;
   }
fputs("Calculating output parameters.\n",stderr);
switch (err = CalcOutParams(&header, &param, width, vsf)) {
  case 0: break;
  case ERR_OUTMEM: fputs("Failed to allocate memory for output image.\n", stderr);
       return 4;
  case ERR_HEADER: fputs("Invalid header passed.\n", stderr);
  default: fputs("Unknown error occured == BUG in program.\n", stderr);
           return 100;
  }
fputs("Converting.\n",stderr);
switch (err = Convert(&header, &param)) {
  case 0: break;
  case ERR_HEADER: fputs("Invalid header passed.\n", stderr);
                   return 2;
  case ERR_COORDINATES: fputs("Coordinates out of range.\n", stderr);
                        return 2;
  default: fputs("Unknown error occured == BUG in program.\n", stderr);
           return 100;
  }
switch (err = Preview(&param)) {
  case 1: fprintf(stderr, "Saving to file: %s\n", outfile);
          switch (err = Save(&param, outfile)) {
            case 0: break;
            case ERR_IOOPEN: ;
            case ERR_IOWRITE: IoErrorReport(err);
                 return 3;
            default: fputs("Unknown error occured == BUG in program.\n", stderr);
                     return 100;
            }
          fputs("Conversion completed.\n",stderr);
          break;
  default: fputs("Conversion aborted.\n",stderr);
           break;
  }
return 0;
}